"""Testing loading of data."""
from sklearn.linear_model import LinearRegression

from feltlabs.config import parse_training_args
from feltlabs.core.data import load_data

csv_data = "1,2,3\n0,2,3\n1,1,4\n0,2,5\n1,1,6"


def test_csv_load(tmp_path):
    # Create CSV files
    for i in range(2):
        file_path = tmp_path / f"{i}"
        with open(file_path, "w") as f:
            f.write(csv_data)

    args_str = f" --data_type csv --input_folder {tmp_path} --aggregation_key fe"
    # Define extra args with different output folder
    args = parse_training_args(args_str.split())
    X, y = load_data(args)

    assert len(X) == len(y), "Size of X and y are not matching"
    assert len(X) == 2 * 5, "Not all data loaded"

    # Try if training works
    model = LinearRegression()
    model.fit(X, y)



"""Module for storing and managing data files."""
from typing import Any, Optional

from feltlabs.core import sklearn_to_json
from feltlabs.core.cryptography import encrypt_nacl
from feltlabs.typing import FileType, Model, PathType


def load_model(filename: FileType) -> Model:
    """Abstraction function for loading models.

    Args:
        filename: filepath or file-like object to load

    Returns:
        model object
    """
    return sklearn_to_json.import_model(filename)


def export_model(model: Model, path: Optional[PathType] = None) -> bytes:
    """Abstraction function for exporting model to file."""
    return sklearn_to_json.export_model(model, path)


def encrypt_model(model: Any, public_key: bytes) -> bytes:
    """Encrypt final model using provided public key.

    Args:
        model: scikit-learn trained model
        public_key: public key to use for encryption

    Returns:
        encrypted model as bytes
    """
    model_bytes = export_model(model)
    encrypted_model = encrypt_nacl(public_key, model_bytes)
    return encrypted_model



"""Testing model aggregation functions."""
import numpy as np
from sklearn.linear_model import LinearRegression

from feltlabs.core.aggregation import (
    aggregate_models,
    random_model,
    remove_noise_models,
    sum_models,
)

X = [[0], [1], [2]]
y = [[0, 1, 2], [1, 2, 3], [0, 2, 4]]


def test_aggregation():
    model1 = LinearRegression()
    model1.fit(X, y[0])

    model2 = LinearRegression()
    model2.fit(X, y[1])

    model3 = LinearRegression()
    model3.fit(X, y[2])

    assert np.allclose(model1.predict(X), y[0])
    assert np.allclose(model2.predict(X), y[1])

    agg_model = aggregate_models([model1, model2])
    assert np.allclose(agg_model.predict(X), [0.5, 1.5, 2.5])

    agg_model = aggregate_models([model1, model3])
    assert np.allclose(agg_model.predict(X), [0, 1.5, 3])

    rand_model1 = random_model(model1, 10)
    enc_model1 = sum_models([model1, rand_model1])
    assert not np.allclose(enc_model1.predict(X), y[0])

    rand_model2 = random_model(model2, 11)
    enc_model2 = sum_models([model2, rand_model2])
    assert not np.allclose(enc_model2.predict(X), y[1])

    tmp_model = aggregate_models([enc_model1, enc_model2])
    final_model = remove_noise_models(tmp_model, [rand_model1, rand_model2])
    assert np.allclose(final_model.predict(X), [0.5, 1.5, 2.5])



"""Module for executing model training."""
from typing import Optional

from feltlabs.config import parse_training_args
from feltlabs.core.aggregation import random_model, sum_models
from feltlabs.core.data import load_data
from feltlabs.core.ocean import save_output
from feltlabs.core.storage import encrypt_model, load_model
from feltlabs.core.training import train_model


def main(args_str: Optional[list[str]] = None, output_name: str = "model"):
    """Main function for exectuting from command line.

    Args:
        args_str: list with string arguments or None if using command line
        output_name: name of output model
    """
    args = parse_training_args(args_str)
    # Load data and model
    model = load_model(args.custom_data_path)
    data = load_data(args)
    # Train model
    model = train_model(model, data)
    # Add randomness to model
    rand_model = random_model(model, args.seed)
    model = sum_models([model, rand_model])
    # Encrypt model using public key of aggregation
    enc_model = encrypt_model(model, args.aggregation_key)

    # Save models into output
    save_output(output_name, enc_model, args)
    print("Training finieshed.")
    return enc_model


if __name__ == "__main__":
    main()



"""Module for special FELT types."""
from pathlib import Path
from typing import Any, Union

# Use this type for all models (TODO: better type)
Model = Any

# File type definiton
FileType = Union[str, Path, bytes]
PathType = Union[str, Path]



"""Testing import/export of sklearn models to json."""
import numpy as np
from sklearn.linear_model import LinearRegression

from feltlabs.core.sklearn_to_json import export_model, import_model

X, y = [[0, 0], [1, 1], [2, 2]], [0, 1, 2]


def test_import_export(tmp_path):
    model = LinearRegression()
    model.fit(X, y)

    file_path = tmp_path / "model.json"

    export_model(model, file_path)
    model_bytes = export_model(model)
    with open(file_path, "rb") as f:
        assert f.read() == model_bytes

    im_model = import_model(file_path)

    assert np.array_equal(im_model.coef_, model.coef_)
    assert np.array_equal(im_model.predict(X), model.predict(X))



"""Module for performing federated averaging of models."""
import copy
import random
from typing import Any, Callable

import numpy as np

from feltlabs.typing import Model

ATTRIBUTE_LIST = ["coef_", "intercept_", "coefs_", "intercepts_"]


def _get_models_params(models: list[Model]) -> dict[str, list[np.ndarray]]:
    """Extract trainable parameters from scikit-learn models.

    Args:
        models: list of scikit-learn models.

    Returns:
        dictionary mapping attributes to list of values numpy arrays extracted from models.
    """
    params = {}
    for param in ATTRIBUTE_LIST:
        params[param] = []
        for model in models:
            if hasattr(model, param):
                params[param].append(getattr(model, param))
            else:
                params.pop(param, None)
                break

    return params


def _get_model_params(model: Model) -> dict[str, np.ndarray]:
    """Extract trainable parameters from scikit-learn model.

    Args:
        model: list of scikit-learn models.

    Returns:
        dictionary mapping attributes to list of values numpy arrays extracted from models.
    """
    params = {}
    for param in ATTRIBUTE_LIST:
        if hasattr(model, param):
            params[param] = getattr(model, param)

    return params


def _set_model_params(model: Model, params: dict[str, np.ndarray]) -> Model:
    """Set new values of trainable params to scikit-learn models.

    Args:
        model: scikit-learn model.
        params: dictinary mapping attributes to numpy arrays.

    Returns:
        scikit-learn model with new values.
    """
    for param, value in params.items():
        setattr(model, param, value)
    return model


def _agg_models_op(
    op: Callable,
    models: list[Model],
) -> Model:
    """Perform aggregation operation on list of scikit-learn models.

    Args:
        op: function to run on model values
        models: list of scikit-learn models

    Returns:
        scikit-learn model with new values based on op
    """
    params = _get_models_params(models)
    new_params = {}
    for param, values in params.items():
        val = op(values)
        new_params[param] = val.astype(values[0].dtype)

    model = _set_model_params(copy.deepcopy(models[0]), new_params)
    return model


def sum_models(models: list[Model]) -> Model:
    """Sum trainable parameters of scikit-learn models.

    Args:
        models: list of scikit-learn models.

    Returns:
        scikit-learn model with new values.
    """
    op = lambda x: np.sum(x, axis=0)
    return _agg_models_op(op, models)


def aggregate_models(models: list[Model]) -> Model:
    """Aggregate trainable parameters of scikit-learn models.

    Args:
        models: list of scikit-learn models.

    Returns:
        scikit-learn model with new values.
    """
    op = lambda x: np.mean(x, axis=0)
    return _agg_models_op(op, models)


def remove_noise_models(main_model: Model, random_models: list[Model]) -> Model:
    """Remove added noise of random models from main scikit-learn model.

    Args:
        model: main scikit-learn model
        models: list of scikit-learn models with random values.

    Returns:
        scikit-learn model with new values.
    """
    op = lambda x: -1 * np.mean(x, axis=0)
    noise_model = _agg_models_op(op, random_models)
    return sum_models([main_model, noise_model])


def _set_seed(seed: int) -> None:
    """Set seed of random generator."""
    np.random.seed(seed)
    random.seed(seed)


def random_model(model: Model, seed: int, min: int = -100, max: int = 100) -> Any:
    """Generate models with random parameters.

    Args:
        model: scikit-learn models.
        seed: seed for randomness generation
        min: minimum value of random number
        max: maximum value of random number

    Returns:
        scikit-learn model with random values.
    """
    _set_seed(seed)

    params = _get_model_params(model)
    new_params = {}
    for param, value in params.items():
        val = (max - min) * np.random.random(value.shape) + min
        new_params[param] = val.astype(val.dtype)

    model = _set_model_params(copy.deepcopy(model), new_params)
    return model



"""Test model download."""
import json

from nacl.public import PrivateKey

from feltlabs.config import parse_aggregation_args
from feltlabs.core.data import load_models

aggregation_key = PrivateKey.from_seed(b"4" * 32)

# Test model url encrypted by above aggregation key
conf_dict = {
    "model_urls": ["https://github.com/FELT-Labs/feltlabs.py/files/9388706/model.txt"]
}


def test_training(tmp_path):
    input_folder = tmp_path / "input"
    output_folder = tmp_path / "output"

    input_folder.mkdir()
    output_folder.mkdir()

    # Create custom data file (containing model URLs)
    with open(input_folder / "algoCustomData.json", "w") as f:
        json.dump(conf_dict, f)

    args_str = f"--output_folder {output_folder}"
    args_str += f" --input_folder {input_folder}"
    args_str += f" --private_key {bytes(aggregation_key).hex()}"
    args_str += f" --download_models"

    # Get aggregation arguments
    args = parse_aggregation_args(args_str.split())
    # Download models from urls
    models = load_models(args)
    assert models[0].predict([[1]]) == 1



"""Module for loading data and models."""
import json
from typing import Any

import numpy as np
import requests

from feltlabs.config import AggregationConfig, TrainingConfig
from feltlabs.core.cryptography import decrypt_nacl
from feltlabs.core.ocean import get_dataset_files
from feltlabs.core.storage import load_model


# TODO: Add model type
def load_data(config: TrainingConfig) -> tuple[np.ndarray, np.ndarray]:
    """Load data and return them in (X, y) format."""
    if config.data_type == "test":
        # Demo data for testing
        X = np.random.rand(100, 10)
        y = np.random.rand(100)
        return X, y
    else:
        files = get_dataset_files(config)
        if config.data_type == "csv":
            X, y = [], []
            for f in files:
                data = np.genfromtxt(f, delimiter=",")
                X.append(data[:, :-1])
                y.append(data[:, -1])

            return np.concatenate(X, axis=0), np.concatenate(y, axis=0)

    raise Exception("No data loaded.")


def load_models(config: AggregationConfig) -> list[Any]:
    """Load models for aggregation.
    It either download models from urls or load them from input path.
    """
    data_array = []
    if config.download_models:
        with config.custom_data_path.open("r") as f:
            conf = json.load(f)

        for url in conf["model_urls"]:
            res = requests.get(url)
            data_array.append(res.content)

    else:
        files = get_dataset_files(config)
        # Decrypt models using private key
        for file_path in files:
            with open(file_path, "rb") as f:
                data_array.append(f.read())

    models = []
    for val in data_array:
        data = decrypt_nacl(config.private_key, val)
        models.append(load_model(data))

    return models



"""Function specific for Ocean's compute-to-data environment."""
from pathlib import Path

from feltlabs.config import OceanConfig


def get_dataset_files(config: OceanConfig) -> list[Path]:
    """Get all file paths provided in Ocean's compute job environment.

    Args:
        config: ocean config containing output path

    Returns:
        list of file paths
    """
    files = []
    for path in config.input_folder.glob("**/*"):
        if path.is_file() and path.name != config.custom_data:
            files.append(path)
    return files


def save_output(name: str, data: bytes, config: OceanConfig) -> None:
    """Save data to output folder by given name.

    Args:
        name: name of file to write
        data: data as bytes to store in the file
        config: ocean config containing output path
    """
    with open(config.output_folder / name, "wb+") as f:
        f.write(data)



"""Module for managing encryption/decryption using keys."""
from base64 import a85decode, a85encode

from nacl.public import Box, PrivateKey, PublicKey


def export_public_key(private_key: bytes) -> bytes:
    """Export public key for contract join request.

    Args:
        private_key: bytes representing private key

    Returns:
        32 bytes representing public key
    """
    return bytes(PrivateKey(private_key).public_key)


def encrypt_nacl(public_key: bytes, data: bytes) -> bytes:
    """Encryption function using NaCl box compatible with MetaMask
    For implementation used in MetaMask look into: https://github.com/MetaMask/eth-sig-util

    Args:
        public_key: public key of recipient
        data: message data

    Returns:
        encrypted data
    """
    emph_key = PrivateKey.generate()
    enc_box = Box(emph_key, PublicKey(public_key))
    # Encryption is required to work with MetaMask decryption (requires utf8)
    data = a85encode(data)
    ciphertext = enc_box.encrypt(data)
    return bytes(emph_key.public_key) + ciphertext


def decrypt_nacl(private_key: bytes, data: bytes) -> bytes:
    """Decryption function using NaCl box compatible with MetaMask
    For implementation used in MetaMask look into: https://github.com/MetaMask/eth-sig-util

    Args:
        private_key: private key to decrypt with
        data: encrypted message data

    Returns:
        decrypted data
    """
    emph_key, ciphertext = data[:32], data[32:]
    box = Box(PrivateKey(private_key), PublicKey(emph_key))
    return a85decode(box.decrypt(ciphertext))



"""Test encryption/decription mechanism."""
from nacl.public import PrivateKey

from feltlabs.core.cryptography import decrypt_nacl, encrypt_nacl, export_public_key


def test_cryptography():
    priv_key = PrivateKey.generate()
    priv_key_b = bytes(priv_key)

    data = b"test data"

    pub_key = export_public_key(priv_key_b)
    assert pub_key == bytes(priv_key.public_key)

    ciphertext = encrypt_nacl(pub_key, data)
    plaintext = decrypt_nacl(priv_key_b, ciphertext)

    assert plaintext == b"test data"



"""Module for exporting sklearn models to json."""
import json
from typing import Any, Optional, Union

import numpy as np
from sklearn import linear_model

from feltlabs.core.aggregation import random_model, remove_noise_models
from feltlabs.typing import FileType, Model, PathType

ATTRIBUTE_LIST = ["coef_", "intercept_", "coefs_", "intercepts_", "classes_", "n_iter_"]
SUPPORTED_MODELS = {
    "LogisticRegression": linear_model.LogisticRegression,
    "LinearRegression": linear_model.LinearRegression,
}


def _model_name(model: Model) -> str:
    """Get name of model class."""
    for name, val in SUPPORTED_MODELS.items():
        if isinstance(model, val):
            return name
    raise Exception("Trying to export unsupported model")


def export_model(model: Model, filename: Optional[PathType] = None) -> bytes:
    """Export sklean model to JSON file or return it as bytes.

    Args:
        model: sklearn model
        filename: path to exported file

    Returns:
        bytes of JSON file
    """
    data = {
        "model_name": _model_name(model),
        "init_params": model.get_params(),
        "model_params": {},
    }

    for p in ATTRIBUTE_LIST:
        if hasattr(model, p):
            data["model_params"][p] = getattr(model, p).tolist()

    if filename:
        with open(filename, "w") as f:
            json.dump(data, f)

    return bytes(json.dumps(data), "utf-8")


def import_model(file: FileType) -> Model:
    """Import sklearn model from file.

    Args
        file: path to model file or bytes

    Returns:
        sklearn model object
    """
    if type(file) is bytes:
        data = json.loads(file)
    else:
        with open(file, "r") as f:
            data = json.load(f)

    if data["model_name"] not in SUPPORTED_MODELS:
        raise Exception("Unsupported model type")

    model_class = SUPPORTED_MODELS[data["model_name"]]
    model = model_class(**data.get("init_params", {}))
    for name, values in data.get("model_params", {}).items():
        setattr(model, name, np.array(values))

    # Substract random models (generated from seeds) from loaded model
    if "seeds" in data:
        rand_models = [random_model(model, s) for s in data.get("seeds", [])]
        model = remove_noise_models(model, rand_models)

    return model



"""Module for handling final models - loading model, exporting model to pickle format."""
import argparse
import pickle
from pathlib import Path

from feltlabs.core.sklearn_to_json import import_model
from feltlabs.typing import Model, PathType


def load_model(path: PathType) -> Model:
    """Load model from json file (intended for use in 3rd party programs).

    Args:
        path: path to json file containing model produced by FELT labs.

    Returns:
        scikit-learn model
    """
    return import_model(path)


def main():
    """Load model from JSON file and save it as pickle file based on arguments."""

    parser = argparse.ArgumentParser(
        description="Script for training models, possible to execute from command line."
    )
    parser.add_argument(
        "--input",
        type=Path,
        help="Path to input model file in JSON format.",
    )
    parser.add_argument(
        "--output",
        type=Path,
        help="Path to output file.",
    )
    args = parser.parse_args()

    model = load_model(args.input)
    pickle.dump(model, args.output)


if __name__ == "__main__":
    main()



"""Module for basic interactions with user."""


def yes_no_prompt(prompt: str, default: bool = False) -> bool:
    """Print yes/no question and check for answer.

    Args:
        prompt: yes/no question to ask
        default: default behavior if answer not matched

    Returns:
        true if user answered yes, false otherwise
    """
    x = input(f"{prompt} {'[Y/n]' if default else '[y/N]'}: ")

    if x.lower().strip() in ["y", "ye", "yes"]:
        return True

    if x.lower().strip() in ["n", "no"]:
        return False

    return default



"""Module for aggregating outputs of local training."""
from typing import Optional

from feltlabs.config import parse_aggregation_args
from feltlabs.core.aggregation import aggregate_models
from feltlabs.core.data import load_models
from feltlabs.core.ocean import save_output
from feltlabs.core.storage import encrypt_model, export_model


def main(args_str: Optional[list[str]] = None, output_name: str = "model"):
    """Main function for exectuting from command line.

    Args:
        args_str: list with string arguments or None if using command line
        output_name: name of output model
    """
    args = parse_aggregation_args(args_str)
    # Load models
    models = load_models(args)
    # Aggregate
    model = aggregate_models(models)
    # Encrypt final model using scientist public key if provided
    if args.public_key:
        model = encrypt_model(model, args.public_key)
    else:
        model = export_model(model)
    # Save model (bytes) into output
    save_output(output_name, model, args)
    print("Aggregation finieshed.")
    return model


if __name__ == "__main__":
    main()



"""FELT python package intended for running federated learning on Ocean protocol.

This code is intended to work closely with Ocean protocol. Algorithms from this code
should run on ocean provider. Training local models and aggregating them into global
model.

Entry commands:

```bash
felt-train
felt-aggregate
```
"""
import re
from pathlib import Path

from setuptools import find_packages, setup

PATH = Path(__file__).parent.absolute()
DOCLINES = (__doc__ or "").split("\n")

CLASSIFIERS = """\
Intended Audience :: Science/Research
Intended Audience :: Developers
Programming Language :: Python
Programming Language :: Python :: 3
Programming Language :: Python :: 3.8
Programming Language :: Python :: 3.9
Programming Language :: Python :: 3.10
Programming Language :: Python :: 3 :: Only
Topic :: Scientific/Engineering
"""


def parse_requirements(file_name):
    """
    from:
        http://cburgmer.posterous.com/pip-requirementstxt-and-setuppy
    """
    requirements = []
    with open(file_name, "r") as f:
        for line in f:
            if re.match(r"(\s*#)|(\s*$)", line):
                continue
            if re.match(r"\s*-e\s+", line):
                requirements.append(
                    re.sub(r"\s*-e\s+.*#egg=(.*)$", r"\1", line).strip()
                )
            elif re.match(r"\s*-f\s+", line):
                pass
            else:
                requirements.append(line.strip())
    return requirements


requirements = parse_requirements(PATH / "requirements.txt")


setup(
    name="feltlabs",
    version="0.2.7",
    packages=find_packages(),
    maintainer="FELT Labs",
    maintainer_email="info@bretahajek.com",
    description=DOCLINES[0],
    long_description="\n".join(DOCLINES[2:]),
    long_description_content_type="text/markdown",
    keywords=["Federated Learning", "Web3", "Machine Learning"],
    url="https://feltlabs.ai/",
    author="FELT Labs",
    project_urls={
        "Bug Tracker": "https://github.com/FELT-Labs/feltlabs.py/issues",
        "Documentation": "https://docs.feltlabs.ai/",
        "Source Code": "https://github.com/FELT-Labs/feltlabs.py",
    },
    license="GPL-3.0 License",
    classifiers=[_f for _f in CLASSIFIERS.split("\n") if _f],
    platforms=["Windows", "Linux", "Solaris", "Mac OS-X", "Unix"],
    python_requires=">=3.8",
    install_requires=requirements,
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "felt-train = feltlabs.algorithm.train:main",
            "felt-aggregate = feltlabs.algorithm.aggregate:main",
            "felt-export = feltlabs.model:main",
        ],
    },
)



"""Module for training models."""
from typing import Any

import numpy as np

from feltlabs.typing import Model


def train_model(
    model: Model,
    data: tuple[np.ndarray, np.ndarray],
) -> Any:
    """Universal model training function, it starts training depending on selected type.

    Args:
        model: initial model object with fit() method
        data: tuple[X, Y] or string depending on training type

    Returns:
        new model object with trained weights
    """
    model.fit(data[0], data[1])
    return model



"""Test training process."""
import json

from nacl.public import PrivateKey

from feltlabs.algorithm import aggregate, train
from feltlabs.config import parse_training_args
from feltlabs.core.aggregation import random_model, remove_noise_models
from feltlabs.core.cryptography import decrypt_nacl
from feltlabs.core.data import load_data
from feltlabs.core.ocean import save_output
from feltlabs.core.storage import load_model

# data = np.array([[0, 0], [1, 1], [2, 2]]), np.array([0, 1, 2])

aggregation_key = PrivateKey.generate()
scientist_key = PrivateKey.generate()

model_def = {"model_name": "LinearRegression"}


def test_training(tmp_path):
    input_folder = tmp_path / "input"
    output_folder = tmp_path / "output"
    output_folder2 = tmp_path / "output2"

    input_folder.mkdir()
    output_folder.mkdir()
    output_folder2.mkdir()

    # Create custom data file (containing model definition)
    with open(input_folder / "algoCustomData.json", "w") as f:
        json.dump(model_def, f)

    enc_models, seeds = [], []

    ### Training section ###
    args_str = f"--output_folder {output_folder}"
    args_str += f" --input_folder {input_folder}"
    args_str += f" --aggregation_key {bytes(aggregation_key.public_key).hex()}"

    # Define extra args with different output folder
    args = parse_training_args(args_str.split())
    args.output_folder = output_folder2

    for i in range(2):
        args_str_final = f"{args_str} --seed {i}"
        enc_model = train.main(args_str_final.split(), f"model_{i}")

        save_output(f"model_{i}", enc_model, args)

        enc_models.append(enc_model)
        seeds.append(i)

    ### Aggregation section ###
    args_str = f"--output_folder {output_folder2}"
    args_str += f" --input_folder {output_folder2}"
    args_str += f" --private_key {bytes(aggregation_key).hex()}"

    enc_final_model = aggregate.main(args_str.split(), "final_model")

    ### Test final results ###
    final_model = load_model(enc_final_model)
    rand_models = [random_model(final_model, s) for s in seeds]
    model = remove_noise_models(final_model, rand_models)

    # Predict
    data = load_data(args)
    model.predict(data[0])



"""Module for loading data provider (Node) config."""
import argparse
import json
from pathlib import Path
from typing import Optional, cast

OUTPUT_FOLDER = Path("/data/outputs")
INPUT_FOLDER = Path("/data/inputs/")
CUSTOM_DATA = "algoCustomData.json"


class OceanConfig:
    input_folder: Path
    output_folder: Path
    custom_data_path: Path
    custom_data: str


class TrainingConfig(OceanConfig):
    aggregation_key: bytes
    data_type: str
    seed: int


class AggregationConfig(OceanConfig):
    private_key: bytes
    public_key: bytes
    download_models: bool


def _add_ocean_config(config: OceanConfig) -> dict[str, str]:
    """Load json file containing algorithm's custom data and add them to config.

    Args:
        config: ocean config containing output path

    Returns:
        dict representing loaded JSON file
    """
    config.custom_data_path = config.input_folder / config.custom_data
    if not config.custom_data_path.exists():
        return {}

    with config.custom_data_path.open("r") as f:
        return json.load(f)


def _ocean_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Add arguments for parsing ocean path configs."""
    parser.add_argument(
        "--output_folder",
        type=Path,
        default=OUTPUT_FOLDER,
        help="Folder for storing outputs.",
    )
    parser.add_argument(
        "--input_folder",
        type=Path,
        default=INPUT_FOLDER,
        help="Folder containing input data.",
    )
    parser.add_argument(
        "--custom_data",
        type=str,
        default=CUSTOM_DATA,
        help="Name of custom data file",
    )
    return parser


def parse_training_args(args_str: Optional[list[str]] = None) -> TrainingConfig:
    """Parse and partially validate arguments form command line.
    Arguments are parsed from string args_str or command line if args_str is None

    Args:
        args_str: list with string arguments or None if using command line

    Returns:
        Parsed args object
    """
    parser = argparse.ArgumentParser(
        description="Script for training models, possible to execute from command line."
    )
    parser = _ocean_parser(parser)
    parser.add_argument(
        "--aggregation_key",
        type=str,
        help="Public key used for encrypting local model for aggregation algorithm.",
    )
    parser.add_argument(
        "--data_type",
        type=str,
        default="test",
        choices=["test", "csv"],
        help="Select type of data. For csv last column is used as Y.",
    )
    parser.add_argument(
        "--seed",
        type=int,
        default=42,
        help="Seed for randomness generation. It should be different for every run.",
    )
    args = parser.parse_args(args_str)

    conf = _add_ocean_config(cast(OceanConfig, args))
    args.data_type = conf["data_type"] if "data_type" in conf else args.data_type
    args.seed = conf["seed"] if "seed" in conf else args.seed

    args.aggregation_key = bytes.fromhex(args.aggregation_key)

    return cast(TrainingConfig, args)


def parse_aggregation_args(args_str: Optional[list[str]] = None) -> AggregationConfig:
    """Parse and partially validate arguments form command line.
    Arguments are parsed from string args_str or command line if args_str is None

    Args:
        args_str: list with string arguments or None if using command line

    Returns:
        Parsed args object
    """
    parser = argparse.ArgumentParser(
        description="Script for training models, possible to execute from command line."
    )
    parser = _ocean_parser(parser)
    parser.add_argument(
        "--private_key",
        type=str,
        help="Private key specified for aggregation algorithm used for decrypting models.",
    )
    parser.add_argument(
        "--public_key",
        type=str,
        help="Public key used for encrypting final model for scientis.",
        default=None,
    )
    parser.add_argument(
        "--download_models",
        help="If true, models will be donwloaded from provided URLs",
        action="store_true",
        default=False,
    )
    args = parser.parse_args(args_str)

    conf = _add_ocean_config(cast(OceanConfig, args))
    args.public_key = conf["public_key"] if "public_key" in conf else args.public_key

    args.private_key = bytes.fromhex(args.private_key)
    if args.public_key:
        args.public_key = bytes.fromhex(args.public_key)

    return cast(AggregationConfig, args)
